# Jags-Ymet-XmetMulti-Mrobust.R 
# Accompanies the book:
#   Kruschke, J. K. (2015). Doing Bayesian Data Analysis, Second Edition: 
#   A Tutorial with R, JAGS, and Stan. Academic Press / Elsevier.

#source("DBDA2E-utilities.R")

#===============================================================================

genMCMCYMetXMetThreeMRobustSubj = function( data , xName="x" , yName="y", subjName="subject",
                    numSavedSteps=10000 , thinSteps=1 , saveName=NULL  ,
                    runjagsMethod="parallel", 
                    nChains=4) { 
  #-----------------------------------------------------------------------------
  # THE DATA.
  y = data[,yName]
  x = as.matrix(data[,xName],ncol=length(xName))
  # Do some checking that data make sense:
  if ( any( !is.finite(y) ) ) { stop("All y values must be finite.") }
  if ( any( !is.finite(x) ) ) { stop("All x values must be finite.") }
  cat("\nCORRELATION MATRIX OF PREDICTORS:\n ")
  show( round(cor(x),3) )
  cat("\n")
  flush.console()
  Nsubj = length(unique(data[,subjName]))
  # Specify the data in a list, for later shipment to JAGS:
  dataList = list(
    x = x ,
    y = y ,
    s = as.numeric(data[,subjName]),
    Nx = dim(x)[2] ,
    Ntotal = dim(x)[1],
    Nsubj = Nsubj
  )
  #-----------------------------------------------------------------------------
  # THE MODEL.
  modelString = "
  # Standardize the data:
  data {
    ym <- mean(y)
    ysd <- sd(y)
    for ( i in 1:Ntotal ) {
      zy[i] <- ( y[i] - ym ) / ysd
    }
    for ( j in 1:Nx ) {
      xm[j]  <- mean(x[,j])
      xsd[j] <-   sd(x[,j])
      for ( i in 1:Ntotal ) {
        zx[i,j] <- ( x[i,j] - xm[j] ) / xsd[j]
      }
    }
  }
  # Specify the model for standardized data:
  model {
    for ( i in 1:Ntotal ) {
      zy[i] ~ dt(zbeta0[s[i]] + zbeta1[s[i]]*zx[i,1]
              + zbeta2[s[i]]*zx[i,2] + zbeta3[s[i]]*zx[i,3],
              1/zsigma^2 , nu)
    }
    for (j in 1:Nsubj){
      zbeta0[j] ~ dnorm(zbeta0mu , 1/(zbeta0sigma)^2)  
      zbeta1[j] ~ dnorm(zbeta1mu , 1/(zbeta1sigma)^2)
      zbeta2[j] ~ dnorm(zbeta2mu , 1/(zbeta2sigma)^2)
      zbeta3[j] ~ dnorm(zbeta3mu , 1/(zbeta3sigma)^2)
    }

    zbeta0mu ~ dnorm( 0 , 1/(10)^2)
    zbeta1mu ~ dnorm( 0 , 1/(10)^2)
    zbeta2mu ~ dnorm( 0 , 1/(10)^2)
    zbeta3mu ~ dnorm( 0 , 1/(10)^2)
    zsigma ~ dunif( 1.0E-3 , 1.0E+3)
    zbeta0sigma ~ dunif( 1.0E-3 , 1.0E+3)
    zbeta1sigma ~ dunif( 1.0E-3 , 1.0E+3) 
    zbeta2sigma ~ dunif( 1.0E-3 , 1.0E+3)
    zbeta3sigma ~ dunif( 1.0E-3 , 1.0E+3)
    #zsigma ~ dunif(1.0E-5 , 1.0E+1)
    nu <- nuMinusOne+1
    nuMinusOne ~ dexp(1/29.0)
    # Transform to original scale:

    for ( j in 1:Nsubj ) {
      beta1[j] <- zbeta1[j] * ysd / xsd[1]
      beta2[j] <- zbeta2[j] * ysd / xsd[2]
      beta3[j] <- zbeta3[j] * ysd / xsd[3]  
      beta0[j] <- zbeta0[j] * ysd  + ym - (zbeta1[j] * xm[1]/ xsd[1] + zbeta2[j] * xm[2]/ xsd[2] + zbeta3[j] * xm[3]/ xsd[3]) * ysd 
    }
    beta1mu <- zbeta1mu * ysd / xsd[1]
    beta2mu <- zbeta2mu * ysd / xsd[2]
    beta3mu <- zbeta3mu * ysd / xsd[3]  
    beta0mu <- zbeta0mu * ysd  + ym - (zbeta1mu * xm[1]/ xsd[1] + zbeta2mu * xm[2]/ xsd[2] + zbeta3mu * xm[3]/ xsd[3]) * ysd 
    sigma <- zsigma * ysd
  }
  " # close quote for modelString
  # Write out modelString to a text file
  #writeLines(modelString , con="TEMPmodel.txt")
  #-----------------------------------------------------------------------------
  # INTIALIZE THE CHAINS.
  # Let JAGS do it...
  initsFunction = function(chain){
      .RNG.seed = c(14, 10, 85, 2017)[chain]
      .RNG.name = c("base::Super-Duper",
                     "base::Wichmann-Hill",
                     "base::Marsaglia-Multicarry",
                     "base::Mersenne-Twister")[chain]
      return(list(.RNG.seed=.RNG.seed,
                  .RNG.name=.RNG.name))
 }
  
  
  #-----------------------------------------------------------------------------
  # RUN THE CHAINS
  parameters = c( "beta0" , "beta1",  "beta2", "beta3", "sigma", 
                  "zbeta0" , "zbeta1", "zbeta2", "zbeta3", "zsigma", "nu",
                 "zbeta0mu" , "zbeta1mu", "zbeta2mu", "zbeta3mu",
                 "beta0mu", "beta1mu",  "beta2mu", "beta3mu")
  adaptSteps = 1000  # Number of steps to "tune" the samplers
  burnInSteps = 2000
  runJagsOut <- run.jags(method="parallel",
                         model=modelString,
                         monitor=parameters, 
                         data=dataList,  
                         inits=initsFunction, 
                         n.chains=nChains,
                         adapt=adaptSteps,
                         burnin=burnInSteps, 
                         sample=ceiling(numSavedSteps/nChains),
                         thin=thinSteps,
                         summarise=FALSE,
                         plots=FALSE)
  codaSamples = as.mcmc.list( runJagsOut )
  if ( !is.null(saveName) ) {
    save(codaSamples , file=paste(saveName,"Mcmc.Rdata",sep="") )
  }
  return( codaSamples )
} # end function

#===============================================================================

smryMCMCYMetXMetThreeMRobustSubj = function(  codaSamples , 
                      saveName=NULL ) {
  summaryInfo = NULL
  mcmcMat = as.matrix(codaSamples,chains=TRUE)
  paramName = colnames(mcmcMat)
  for ( pName in paramName ) {
    summaryInfo = rbind( summaryInfo , summarizePost( mcmcMat[,pName] ) )
  }
  rownames(summaryInfo) = paramName
  summaryInfo = rbind( summaryInfo , 
                       "log10(nu)" = summarizePost( log10(mcmcMat[,"nu"]) ) )
  if ( !is.null(saveName) ) {
    write.csv( summaryInfo , file=paste(saveName,"SummaryInfo.csv",sep="") )
  }
  return( summaryInfo )
}

